#include <err.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <unistd.h>

#include <openssl/err.h>
#include <openssl/ssl.h>
#include <openssl/x509v3.h>

#include "dpi.h"
#include "io.h"

static void respond_errstr(const char *errstr, const char *fmt, ...) {
	va_list ap;
	int err = errno;
	va_start(ap, fmt);
	dpi_send_header(NULL, "text/plain");
	vprintf(fmt, ap);
	printf(": %s", errstr ? errstr : strerror(err));
	va_end(ap);
}

static void respond_err(const char *fmt, ...) {
	va_list ap;
	int err = errno;
	va_start(ap, fmt);
	dpi_send_header(NULL, "text/plain");
	vprintf(fmt, ap);
	printf(": %s", strerror(err));
	va_end(ap);
}

static const char *ssl_strerror(int err) {
	switch (err) {
		case SSL_ERROR_NONE: return "no error";
		case SSL_ERROR_ZERO_RETURN: return "connection closed";
		case SSL_ERROR_WANT_READ: return "read did not complete";
		case SSL_ERROR_WANT_WRITE: return "write did not complete";
		case SSL_ERROR_WANT_CONNECT: return "connect did not complete";
		case SSL_ERROR_WANT_ACCEPT: return "accept did not complete";
		case SSL_ERROR_WANT_X509_LOOKUP: return "cert application callback set";
		case SSL_ERROR_WANT_ASYNC: return "asynchronous engine is still processing data";
		case SSL_ERROR_WANT_ASYNC_JOB: return "no async jobs available";
		case SSL_ERROR_SYSCALL: return strerror(errno);
		case SSL_ERROR_SSL: ERR_print_errors_fp(stderr); return "protocol error"; //return ERR_error_string(ERR_get_error(), NULL);
		default: return "unknown error";
	}
}

static void respond_ssl_err(const SSL *ssl, int rc, const char *fmt, ...) {
	va_list ap;
	int err = errno;
	va_start(ap, fmt);
	dpi_send_header(NULL, "text/plain");
	vprintf(fmt, ap);
	va_end(ap);
	rc = SSL_get_error(ssl, rc);
	if (rc == SSL_ERROR_SSL) {
		printf(": ");
		ERR_print_errors_fp(stdout);
	} else {
		printf(": %s", ssl_strerror(rc));
	}
}

static int parse_port(const char *url, char *port, size_t port_len) {
	while (port_len > 0 && *url != '\0' && *url != '/') {
		*port++ = *url++;
		port_len--;
	}
	if (port_len <= 0) {
		errno = EMSGSIZE;
		return -1;
	}
	*port = '\0';
	return 0;
}

static int parse_url(const char *url, char *host, size_t host_len, char *port, size_t port_len, char *path, size_t path_len) {
	if (!strncmp(url, "gemini:", 7)) url += 7;
	if (!strncmp(url, "//", 2)) url += 2;

	while (host_len > 0 && *url != '\0' && *url != '/' && *url != ':') {
		*host++ = *url++;
		host_len--;
	}
	if (host_len <= 0) {
		errno = EMSGSIZE;
		return -1;
	}
	*host = '\0';
	if (*url == ':') return parse_port(url + 1, port, port_len);
	*port = '\0';
	return 0;
}

static int putchar_htmlenc(char c) {
	switch (c) {
		case '<': return printf("&lt;");
		case '>': return printf("&gt;");
		case '"': return printf("&quot;");
		default: return putchar(c);
	}
}

static void print_htmlenc(const char *str) {
	char c;
	while ((c = *str++)) putchar_htmlenc(c);
}

static int read_status(SSL *ssl) {
	int status = 0;
	char c;
	int rc;
	while ((rc = SSL_read(ssl, &c, 1)) == 1) {
		if (c >= '0' && c <= '9') {
			status = 10*status + (c - '0');
		} else if (c == ' ' || c == '\t') {
			return status;
		}
	}
	return rc;
}

static int read_meta(SSL *ssl, char *buf, size_t *lenp) {
	size_t len = *lenp;
	size_t i = 0;
	char c;
	int rc;

	// ignore initial whitespace
	while (i < len) {
		rc = SSL_read(ssl, &c, 1);
		if (rc < 1) return rc;
		if (c == ' ' || c == '\t') {
			continue;
		} else {
			buf[i++] = c;
			break;
		}
	}

	// read characters until newline or CRLF
	while (i < len) {
		rc = SSL_read(ssl, &c, 1);
		if (rc < 1) return rc;
		if (c == '\n') {
			if (buf[i-1] == '\r') i--;
			break;
		} else {
			buf[i++] = c;
		}
	}

	if (i < len) {
		buf[i++] = '\0';
	}

	*lenp = i;
	return 0;
}

static int read_gemini_line_rest(SSL *ssl) {
	int rc;
	char c;
	while (1) {
		rc = SSL_read(ssl, &c, 1);
		if (rc < 1) return rc;
		if (c == '\r') continue;
		putchar(c);
		if (c == '\n') return 1;
	}
	return rc;
}

static int read_gemini_line(SSL *ssl) {
	int rc;
	char c;
	char href[4096];
	size_t href_len = 0;

	rc = SSL_read(ssl, &c, 1);
	if (rc != 1) return rc;
	if (c == '\n') {
		putchar(c);
		return 1;
	}
	if (c != '=') {
		putchar(c);
		return read_gemini_line_rest(ssl);
	}

	rc = SSL_read(ssl, &c, 1);
	if (rc != 1) return rc;
	if (c != '>') {
		putchar(c);
		return read_gemini_line_rest(ssl);
	}

	do {
		rc = SSL_read(ssl, &c, 1);
		if (rc != 1) return rc;
	} while (c == ' ' || c == '\t');
	href[href_len++] = c;

	while (href_len < sizeof(href)) {
		rc = SSL_read(ssl, &c, 1);
		if (rc != 1) return rc;
		if (c == ' ' || c == '\t' || c == '\r' || c == '\n') break;
		href[href_len++] = c;
	}
	href[href_len] = '\0';
	printf("=> <a href=\"");
	print_htmlenc(href);
	printf("\">");
	print_htmlenc(href);
	printf("</a>");
	putchar(c);
	if (c == '\n') return 1;

	return read_gemini_line_rest(ssl);
}

static void respond_gemini(SSL *ssl, const char *gemini_meta) {
	unsigned long error;
	int rc;
	dpi_send_header(NULL, "text/html");
	printf("<!doctype html><pre>");
	while ((rc = read_gemini_line(ssl)) == 1);
	printf("</pre>");
	error = SSL_get_error(ssl, rc);
	if (error == SSL_ERROR_ZERO_RETURN || error == SSL_ERROR_NONE) return;
	printf("SSL Error: <pre>%s</pre>", ssl_strerror(error));
}

static void respond_success(SSL *ssl, const char *meta) {
	unsigned long error;
	int rc;
	char buf[4096];
	size_t len;

	if (!strncmp(meta, "text/gemini", 11)) return respond_gemini(ssl, meta + 11);

	dpi_send_header(NULL, meta);
	while (1) {
		len = SSL_pending(ssl);
		if (len == 0) len = 1;
		if (len > sizeof(buf)) len = sizeof(buf);
		len = SSL_read(ssl, buf, len);
		if (len < 1) break;
		len = fwrite(buf, 1, len, stdout);
		if (len < 0) errx(1, "failed to print response");
	}
	error = SSL_get_error(ssl, len);
	if (error == SSL_ERROR_ZERO_RETURN || error == SSL_ERROR_NONE) return;
	printf("SSL Error: %s", ssl_strerror(error));
}

static void respond_redirect(SSL *ssl, int status, const char *meta) {
	dpi_send_header(NULL, "text/html");
	printf("<!doctype html><html><head><title>Redirect to ");
	print_htmlenc(meta);
	printf("</title></head><body><h3>Redirect to <a href=\"");
	print_htmlenc(meta);
	printf("\">");
	print_htmlenc(meta);
	printf("</a></h3></body></html>");
}

static void respond_status_not_supported(SSL *ssl, int status, const char *meta) {
	dpi_send_header(NULL, "text/plain");
	printf("Status not supported.");
}

static void respond(const char *url) {
	int rc;
	const char *errstr = NULL;
	char host[1024];
	char port[1024];
	char path[4096];
	int s;
	SSL *ssl;
	SSL_CTX *ssl_ctx;
	X509_VERIFY_PARAM *param;

	rc = parse_url(url, host, sizeof(host), port, sizeof(port), path, sizeof(path));
	if (rc < 0) return respond_err("parse_url");
	if (!*host) strcpy(host, "127.0.0.1");
	if (!*port) strcpy(port, "1965");

	SSL_load_error_strings();

	ssl_ctx = SSL_CTX_new(TLS_client_method());
	if (ssl_ctx == NULL) respond_err("failed to create ssl context");
	(void)SSL_CTX_set_mode(ssl_ctx, SSL_MODE_AUTO_RETRY);

	ssl = SSL_new(ssl_ctx);
	if (ssl_ctx == NULL) respond_err("failed to create ssl object");

	// warnx("connecting to %s: %s", host, port);
	s = tcp_connect(host, port, &errstr);
	if (s < 0) return respond_errstr(errstr, "tcp_connect");


	rc = SSL_set_fd(ssl, s);
	if (rc < 0) return respond_ssl_err(ssl, rc, "failed to set SSL fd");

	param = SSL_get0_param(ssl);
	X509_VERIFY_PARAM_set_hostflags(param, X509_CHECK_FLAG_NO_PARTIAL_WILDCARDS);
	if (!X509_VERIFY_PARAM_set1_host(param, host, strlen(host))) {
		return respond_ssl_err(ssl, rc, "failed to set SSL host");
	}

	// SSL_set_verify(ssl, SSL_VERIFY_PEER, NULL);

	rc = SSL_connect(ssl);
	if (rc < 0) return respond_ssl_err(ssl, rc, "SSL_connect");

	rc = SSL_write(ssl, url, strlen(url));
	if (rc < 0) return respond_ssl_err(ssl, rc, "write request");
	rc = SSL_write(ssl, "\r\n", 2);
	if (rc < 0) return respond_ssl_err(ssl, rc, "write request end");

	// read status+meta line
	int status = read_status(ssl);
	if (status < 0) return respond_ssl_err(ssl, status, "read status");

	char meta[1024];
	size_t meta_len = sizeof(meta);
	rc = read_meta(ssl, meta, &meta_len);
	if (rc < 0) return respond_ssl_err(ssl, rc, "read meta");
	
	// fprintf(stderr, "status %d meta %zu %.*s\n", status, meta_len, (int)meta_len, meta);

	if (status >= 20 && status < 30) respond_success(ssl, meta);
	else if (status >= 30 && status < 40) respond_redirect(ssl, status, meta);
	else respond_status_not_supported(ssl, status, meta);

	rc = SSL_shutdown(ssl);
	// if (rc == 0) rc = SSL_shutdown(ssl);
	// if (rc < 0) return respond_ssl_err(ssl, rc, "SSL_shutdown");
	close(s);
}

int main() {
	char url[1024];
	// setvbuf(stdout, NULL, _IONBF, BUFSIZ);
	dpi_read_request(url, sizeof(url));
	respond(url);
}
